mod bst;
mod sort;
mod stack;

use bst::BinarySearchTree;
use clap::{Args, Parser, Subcommand};
use rand::Rng;
use sort::bucket_sort;
use stack::Stack;

#[derive(Parser, Debug)]
struct Cli {
    #[command(subcommand)]
    command: Option<Command>,
}

#[derive(Subcommand, Debug)]
enum Command {
    Default,
    DoRandom(DoRandomArgs),
}

#[derive(Args, Debug)]
struct DoRandomArgs {
    #[arg(long)]
    min: i32,

    #[arg(long)]
    max: i32,
}

struct Greeter {
    greeting: &'static str,
}

impl Greeter {
    fn new(greeting: &'static str) -> Greeter {
        Greeter { greeting }
    }

    pub fn greet(&self) {
        println!("{}", self.greeting);
    }
}

extern "C" {
    fn print_num(num: i32);
    fn print_a_string(s: *const u8, len: usize);
}

fn do_default() {
    Greeter::new("Hello, Jesse!").greet();

    // Calling some C

    let original_s = "Hello from C-Rust land!";
    unsafe {
        print_a_string(original_s.as_ptr(), original_s.len());
    }

    // Testing out stack

    let mut my_stack = Stack::new();
    my_stack.push(Greeter::new("Hello"));
    my_stack.push(Greeter::new("Beautiful"));
    my_stack.push(Greeter::new("World"));

    for greeter in my_stack {
        greeter.greet();
    }

    // BST

    let mut bst = BinarySearchTree::new();
    bst.insert(1);
    bst.insert(2);
    bst.insert(3);

    println!("{:?}", bst);

    bst.insert(0);
    println!("{:?}", bst);
    println!("find result: {:?}", bst.find(0));

    // calling some C
    let find_result = bst.find(3).unwrap_or(-1);
    unsafe {
        print_num(find_result);
    }

    // linear bucket sort
    let items = [3, 2, 1, 1, 2, 3, 4];
    println!("before sort: {:?}", items);

    let sorted = bucket_sort(&items);
    println!("sorted: {:?}", sorted);

    let some_chars = ['a', 'b', 'c', 'c', 'a', 'b', 'd'];
    let sorted_chars = bucket_sort(&some_chars);
    println!("sorted chars: {:?}", sorted_chars);
}

fn do_random(args: &DoRandomArgs) {
    let mut rng = rand::thread_rng();
    let mut numbers = [0; 100];
    let mut i = 0;
    while i < numbers.len() {
        numbers[i] = rng.gen_range(args.min..args.max);
        i += 1;
    }
    // println!("before sorting: {:?}", numbers);
    let sorted = bucket_sort(&numbers);
    println!("sorted: {:?}", sorted);
}

fn main() {
    let cli = Cli::parse();
    if cli.command.is_none() {
        do_default()
    } else {
        match cli.command.unwrap() {
            Command::Default => {
                do_default();
            }
            Command::DoRandom(args) => {
                do_random(&args);
            }
        }
    }
}
